from __future__ import division

import logging
# logging.basicConfig(level=logging.INFO)
import math

import pandas as pd
from scipy.stats import f


def formatval(v):
    return v if (isinstance(v, str) or int(v) != v) else int(v)


def _plot(
    modelinstance,
    comps=['PC1', 'PC2'],
    classlevels=None,
    markers=None,
    classcolors=None,
    msize=100,
    figsize=(12, 8),
    plotpred=True,
    predsize=None,
    predlevels=None,
    predcolors=None,
    predmarkers=None,
    labels=None,
    predlabels=None,
    textsize=10,
    color='#555555'
):
    """Plot method for plotting scores, with optional classes and predictions"""
    if not markers:
        markers = ['s', '^', 'v', 'o', '<', '>', 'D', 'p']
    if classlevels:
        if not classcolors:
            classcolors = [str(f) for f in pd.np.linspace(0, 1, len(classlevels)+1)[1:]]
        ax = modelinstance.scores.xs(1, level=classlevels[0]).plot(
            kind='scatter',
            x=comps[0], y=comps[1], figsize=figsize, s=msize, zorder=3,
            marker=markers[0], edgecolor='black', linewidth='1', c=classcolors[0])
        for i, lev in enumerate(classlevels[1:]):
            modelinstance.scores.xs(1, level=lev).plot(
                kind='scatter',
                x=comps[0], y=comps[1], s=msize, zorder=3,
                marker=markers[i+1], c=classcolors[i+1],
                edgecolor='black', linewidth='1', ax=ax, grid=True
            )
    else:
        ax = modelinstance.scores.plot(
            kind='scatter',
            x=comps[0], y=comps[1], figsize=figsize, s=msize, zorder=3,
            marker=markers[0], edgecolor='black', linewidth='1', c=color, grid=True)
    el = simpleEllipse(modelinstance.scores[comps[0]], modelinstance.scores[comps[1]], 0.95, 200)
    if labels:
        modelinstance.scores.reset_index(labels).apply(lambda row: ax.annotate(
            row[labels], (row[comps[0]], row[comps[1]]),
            xytext=(10, -5),
            textcoords='offset points',
            size=textsize,
            color='black',
            zorder=4
        ), axis=1)
    ax.plot(el[0], el[1], color='black', linewidth=1)
    ax.axvline(x=0, ls='-', color='black', linewidth=1)
    ax.axhline(y=0, ls='-', color='black', linewidth=1)
    if plotpred and hasattr(modelinstance, 'pred'):
        predsize = predsize or msize * 2
        predmarkers = predmarkers or markers
        try:
            predcolors = predcolors or ['C{}'.format(c+1) for c in range(len(predlevels))]
        except TypeError:
            predcolors = ['C1']
        if predlevels:
            for lev in range(len(predlevels)):
                try:
                    modelinstance.pred.xs(1, level=predlevels[lev]).plot(
                        kind='scatter',
                        x=comps[0], y=comps[1], s=predsize, zorder=5,
                        marker=predmarkers[lev], c=predcolors[lev],
                        edgecolor='black', linewidth='1', ax=ax, grid=True
                    )
                    if not (predlabels is None):
                        modelinstance.pred.xs(1, level=predlevels[lev]).reset_index(predlabels).rename(
                            columns={'index': 'level_0'}
                        ).apply(lambda row: ax.annotate(
                            row[
                                'level_{}'.format(predlabels) if 0 == predlabels else predlabels
                            ],
                            (row[comps[0]], row[comps[1]]),
                            xytext=(10, -5),
                            textcoords='offset points',
                            size=textsize * (2 if predsize is None else (predsize / msize)),
                            color='black',
                            zorder=6
                        ), axis=1)
                except (KeyError, ValueError, IndexError):
                    logging.warning(
                        "Tried to plot prediction data for class {}, but failed. "
                        "Probably there are no datapoints with that class.".format(
                            predlevels[lev]
                        ))
                    raise
        else:
            modelinstance.pred.plot(
                kind='scatter',
                x=comps[0], y=comps[1], s=predsize, zorder=6,
                marker=predmarkers[0], c=predcolors[0],
                edgecolor='black', linewidth='1', ax=ax, grid=True
            )
            if not (predlabels is None):
                modelinstance.pred.reset_index(predlabels).rename(
                    columns={'index': 'level_0'}
                ).apply(lambda row: ax.annotate(
                    formatval(row[
                        'level_{}'.format(predlabels) if 0 == predlabels else predlabels
                    ]),
                    (row[comps[0]], row[comps[1]]),
                    xytext=(10, -5),
                    textcoords='offset points',
                    size=textsize * (2 if predsize is None else (predsize / msize)),
                    color='black',
                    zorder=7
                ), axis=1)
    return ax.figure


def _loadingsplot(
    modelinstance,
    comps=['PC1', 'PC2'],
    markers=None,
    color='0.7',
    msize=100,
    figsize=(12, 8),
    showweights=True,
    weightmarkers=None,
    weightcolors=None,
    weightsize=200,
    labels=True,
    textsize=10
):
    """Plot method for plotting loadings"""
    try:
        _loadings = modelinstance.weights
    except AttributeError:
        _loadings = modelinstance.loadings
    ax = _loadings.plot(
        kind='scatter', x=comps[0], y=comps[1], figsize=figsize, s=msize,
        c=color, edgecolor='black', grid=True, zorder=3
    )
    if labels:
        _loadings.apply(lambda row: ax.annotate(
            row.name, (row[comps[0]], row[comps[1]]),
            xytext=(10, -5),
            textcoords='offset points',
            size=textsize,
            color='black',
            zorder=4
        ), axis=1)
    if showweights:
        if weightmarkers is None:
            modelinstance.q.plot(
                kind='scatter', x=comps[0], y=comps[1], s=weightsize,
                c=weightcolors or 'blue', edgecolor='black', grid=True, zorder=5, ax=ax
            )
        else:
            for _m, _c, _x, _y in zip(
                weightmarkers,
                weightcolors or ['blue'] * len(weightmarkers),
                modelinstance.q[comps[0]],
                modelinstance.q[comps[1]]
            ):
                ax.scatter(
                    _x, _y, marker=_m, c=_c, s=weightsize,
                    edgecolor='k', zorder=5
                )
        if labels:
            modelinstance.q.apply(lambda row: ax.annotate(
                row.name, (row[comps[0]], row[comps[1]]),
                xytext=(10, -5),
                textcoords='offset points',
                size=textsize * weightsize / msize,
                color='black',
                zorder=6
            ), axis=1)
    ax.axvline(x=0, ls='-', color='black', linewidth=1, zorder=2.7)
    ax.axhline(y=0, ls='-', color='black', linewidth=1, zorder=2.7)
    return ax.figure


def simpleEllipse(x, y, alfa, length):
    """Helper function to calculate Hotelling's ellipse from scores
    Ported from the bioconductor package pcaMethods, https://doi.org/doi:10.18129/B9.bioc.pcaMethods
    Stacklies W, Redestig H, Scholz M, Walther D and Selbig J (2007).
    “pcaMethods – a Bioconductor package providing PCA methods for incomplete data.”
    Bioinformatics, 23, pp. 1164–1167.
    """
    n = len(x)
    mypi = [i/(length-1)*pd.np.pi*2 for i in range(length)]
    r1 = pd.np.sqrt(x.var() * f.ppf(alfa, 2, n-2) * 2 * (n**2 - 1) / (n * (n - 2)))
    r2 = pd.np.sqrt(y.var() * f.ppf(alfa, 2, n-2) * 2 * (n**2 - 1) / (n * (n - 2)))
    return r1 * pd.np.cos(mypi) + x.mean(), r2 * pd.np.sin(mypi) + y.mean()


class PLS(object):
    """A class for PLS calculated by the NIPALS algorithm.

    Initialize with a Pandas DataFrame or an object that can be turned into a DataFrame
    (e.g. an array or a dict of lists)"""
    def __init__(self, x_df, y_df):
        super(PLS, self).__init__()
        if type(x_df) != pd.core.frame.DataFrame:
            x_df = pd.DataFrame(x_df)
        if type(y_df) != pd.core.frame.DataFrame:
            y_df = pd.DataFrame(y_df)
        # Make sure data is numeric
        self.x_df = x_df.astype('float')
        self.y_df = y_df.astype('float')
        # Check for and remove infs
        if pd.np.isinf(self.x_df).any().any():
            logging.warning("X data contained infinite values, converting to missing values")
            self.x_df.replace([pd.np.inf, -pd.np.inf], pd.np.nan, inplace=True)
        if pd.np.isinf(self.y_df).any().any():
            logging.warning("Y data contained infinite values, converting to missing values")
            self.y_df.replace([pd.np.inf, -pd.np.inf], pd.np.nan, inplace=True)

    def fit(
        self,
        ncomp=None,
        center=True,
        scale=True,
        startcol=None,
        tol=0.000001,
        maxiter=500,
        cv=False
    ):
        """The Fit method, will fit a PLS to the X and Y data"""
        if ncomp is None:
            ncomp = min(self.x_df.shape)
        elif ncomp > min(self.x_df.shape):
            ncomp = min(self.x_df.shape)
            logging.warning(
                'ncomp is larger than the max dimension of the x matrix.\n'
                'fit will only return {} components'.format(ncomp)
            )

        # Convert to np array
        self.x_mat = self.x_df.values
        self.y_mat = self.y_df.values
        if center:
            self.x_mean = pd.np.nanmean(self.x_mat, axis=0)
            self.x_mat = self.x_mat - self.x_mean
            self.y_mean = pd.np.nanmean(self.y_mat, axis=0)
            self.y_mat = self.y_mat - self.y_mean
        if scale:
            self.x_std = pd.np.nanstd(self.x_mat, axis=0, ddof=1)
            self.x_mat = self.x_mat / self.x_std
            self.y_std = pd.np.nanstd(self.y_mat, axis=0, ddof=1)
            self.y_mat = self.y_mat / self.y_std

        TotalSSX = pd.np.nansum(self.x_mat*self.x_mat)
        TotalSSY = pd.np.nansum(self.y_mat*self.y_mat)
        nr, x_nc = self.x_mat.shape
        y_nc = self.y_mat.shape[1]
        # initialize outputs
        # eig = pd.np.empty((ncomp,))
        R2Xcum = pd.np.empty((ncomp,))
        R2Ycum = pd.np.empty((ncomp,))
        PRESS_SS = pd.np.empty((ncomp,))
        loadings = pd.np.empty((x_nc, ncomp))
        scores = pd.np.empty((nr, ncomp))
        u = pd.np.empty((nr, ncomp))
        weights = pd.np.empty((x_nc, ncomp))
        q = pd.np.empty((y_nc, ncomp))
        b = pd.np.empty((ncomp,))

        # NA handling
        x_miss = pd.np.isnan(self.x_mat)
        x_hasna = x_miss.any()
        y_miss = pd.np.isnan(self.y_mat)
        y_hasna = y_miss.any()
        if x_hasna or y_hasna:
            logging.info("Data has NA values")

        if cv:
            if cv is True:
                cv = 7
            cvn = int(pd.np.ceil(nr / cv))
            cvgroups = pd.np.array(range(cvn * cv)).reshape(cvn, cv).T
        else:
            cv = 0
        for comp in range(ncomp):
            PRESS = 0
            for cvround in range(cv + 1):
                if cvround >= cv:
                    # Calculate on full matrix after CV rounds
                    train_x_mat = self.x_mat
                    train_y_mat = self.y_mat
                else:
                    train_x_mat = pd.np.delete(self.x_mat, cvgroups[cvround], 0)
                    train_y_mat = pd.np.delete(self.y_mat, cvgroups[cvround], 0)
                nrt, x_nct = train_x_mat.shape
                y_nct = train_y_mat.shape[1]
                train_x_miss = pd.np.isnan(train_x_mat)
                train_y_miss = pd.np.isnan(train_y_mat)
                # Set u to some column of Y
                if startcol is None:
                    yvar = pd.np.nanvar(self.y_mat, axis=0, ddof=1)
                    startcol_use = pd.np.where(yvar == yvar.max())[0][0]
                else:
                    startcol_use = startcol
                logging.info("PC {}, starting with column {}".format(comp, startcol_use))

                if y_hasna:
                    train_y_mat_0 = pd.np.nan_to_num(train_y_mat)
                    uh = train_y_mat_0[:, startcol_use]
                else:
                    uh = train_y_mat[:, startcol_use]
                th = uh

                if x_hasna:
                    train_x_mat_0 = pd.np.nan_to_num(train_x_mat)

                it = 0
                while True:
                    # X-block weights
                    if x_hasna:
                        U2 = pd.np.repeat(uh*uh, x_nct)
                        U2.shape = (nrt, x_nct)
                        U2[train_x_miss] = 0
                        wh = train_x_mat_0.T.dot(uh) / U2.sum(axis=0)
                    else:
                        wh = train_x_mat.T.dot(uh) / sum(uh*uh)
                    # Normalize
                    wh = wh / math.sqrt(pd.np.nansum(wh*wh))

                    # X-block Scores
                    th_old = th
                    if x_hasna:
                        W2 = pd.np.repeat(wh*wh, nrt)
                        W2.shape = (x_nct, nrt)
                        W2[train_x_miss.T] = 0
                        th = train_x_mat_0.dot(wh) / W2.sum(axis=0)
                    else:
                        th = train_x_mat.dot(wh) / sum(wh*wh)

                    # Y-block weights
                    if y_hasna:
                        T2 = pd.np.repeat(th*th, y_nct)
                        T2.shape = (nrt, y_nct)
                        T2[train_y_miss] = 0
                        qh = train_y_mat_0.T.dot(th) / T2.sum(axis=0)
                    else:
                        qh = train_y_mat.T.dot(th) / sum(th*th)
                    # Normalize
                    # According to Analytica Chimica Acta, 186 (1986) 1-17 this normalization
                    # should be done. However, if so, the results are not the same as th R package
                    # pls plsr method or Evince.
                    # qh = qh / math.sqrt(pd.np.nansum(qh*qh))

                    # Y-block Scores
                    if y_hasna:
                        Q2 = pd.np.repeat(qh*qh, nrt)
                        Q2.shape = (y_nct, nrt)
                        Q2[train_y_miss.T] = 0
                        uh = train_y_mat_0.dot(qh) / Q2.sum(axis=0)
                    else:
                        uh = train_y_mat.dot(qh) / sum(qh*qh)

                    # Check convergence
                    if pd.np.nansum((th-th_old)**2) < tol:
                        break
                    it += 1
                    if it >= maxiter:
                        raise RuntimeError(
                            "Convergence was not reached in {} iterations for component {}".format(
                                maxiter, comp
                            )
                        )

                # Calculate PRESS for this CV
                if cvround < cv:
                    pred_x_mat = self.x_mat[[i for i in cvgroups[cvround] if i < nr]]
                    pred_y_mat = self.y_mat[[i for i in cvgroups[cvround] if i < nr]]
                    pred_x_mat = pd.np.nan_to_num(pred_x_mat)
                    cv_th = pred_x_mat.dot(wh) / sum(wh*wh)
                    cv_bh = sum(uh*th)/sum(th**2)
                    cv_res = pred_y_mat - cv_bh*pd.np.outer(cv_th, qh)
                    cv_res[pd.np.isnan(pred_y_mat)] = 0
                    PRESS += pd.np.sum(cv_res**2)

            PRESS_SS[comp] = PRESS / pd.np.nansum(self.y_mat*self.y_mat)

            # Calculate X loadings and rescale the scores and weights
            if x_hasna:
                T2 = pd.np.repeat(th*th, x_nc)
                T2.shape = (nr, x_nc)
                T2[x_miss] = 0
                ph = train_x_mat_0.T.dot(th) / T2.sum(axis=0)
            else:
                ph = train_x_mat.T.dot(th) / sum(th*th)
            # Normalize
            # According to Analytica Chimica Acta, 186 (1986) 1-17 this normalization
            # should be done. However, if so, the results are not the same as th R package
            # pls plsr method or Evince.
            # pold_len = math.sqrt(pd.np.nansum(ph*ph))
            # ph = ph / pold_len
            # th = th * pold_len
            # wh = wh * pold_len
            loadings[:, comp] = ph
            scores[:, comp] = th
            u[:, comp] = uh
            q[:, comp] = qh
            weights[:, comp] = wh
            bh = sum(uh*th)/sum(th**2)
            b[comp] = bh

            self.x_mat = self.x_mat - pd.np.outer(th, ph)
            self.y_mat = self.y_mat - bh*pd.np.outer(th, qh)

            # Cumulative proportion of variance explained
            R2Xcum[comp] = 1 - (pd.np.nansum(self.x_mat*self.x_mat) / TotalSSX)
            R2Ycum[comp] = 1 - (pd.np.nansum(self.y_mat*self.y_mat) / TotalSSY)

        # "Uncumulate" R2
        self.R2X = pd.np.insert(pd.np.diff(R2Xcum), 0, R2Xcum[0])
        self.R2Y = pd.np.insert(pd.np.diff(R2Ycum), 0, R2Ycum[0])
        self.R2Xcum = pd.Series(R2Xcum, index=["PC{}".format(i+1) for i in range(ncomp)])
        self.R2Ycum = pd.Series(R2Ycum, index=["PC{}".format(i+1) for i in range(ncomp)])

        if cv:
            self.PRESS_SS = pd.Series(PRESS_SS, index=["PC{}".format(i+1) for i in range(ncomp)])
            self.Q2 = 1 - self.PRESS_SS
            self.Q2cum = 1 - pd.np.cumprod(self.PRESS_SS)

        # Convert results to DataFrames
        self.scores = pd.DataFrame(scores, index=self.x_df.index, columns=["PC{}".format(i+1) for i in range(ncomp)])
        self.loadings = pd.DataFrame(loadings, index=self.x_df.columns, columns=["PC{}".format(i+1) for i in range(ncomp)])
        self.u = pd.DataFrame(u, index=self.x_df.index, columns=["PC{}".format(i+1) for i in range(ncomp)])
        self.q = pd.DataFrame(q, index=self.y_df.columns, columns=["PC{}".format(i+1) for i in range(ncomp)])
        self.weights = pd.DataFrame(weights, index=self.x_df.columns, columns=["PC{}".format(i+1) for i in range(ncomp)])
        self.b = pd.Series(b, index=["PC{}".format(i+1) for i in range(ncomp)])
        return True

    def dModY(self):
        """
        Calculates DModY for model, ported from pcaMethods
        (Stacklies W, Redestig H, Scholz M, Walther D and Selbig J (2007).
        “pcaMethods – a Bioconductor package providing PCA methods for incomplete
        data.” Bioinformatics, 23, pp. 1164–1167.)
        Modified to scale with mean
        """
        nr, nc = self.y_mat.shape
        ncomp = self.scores.shape[1]
        A0 = 0 if type(self.y_mean) == int else 1
        ny = pd.np.sqrt(nr / (nr - ncomp - A0))
        F2 = self.y_mat*self.y_mat
        s = pd.np.sqrt(pd.np.nansum(F2, axis=1) / (nc - ncomp)) * ny
        S0 = pd.np.sqrt(pd.np.nansum(F2)/((nr - ncomp - A0) * (nc - ncomp)))
        return s/S0

    def dModX(self):
        """
        Calculates DModX for model, ported from pcaMethods
        (Stacklies W, Redestig H, Scholz M, Walther D and Selbig J (2007).
        “pcaMethods – a Bioconductor package providing PCA methods for incomplete
        data.” Bioinformatics, 23, pp. 1164–1167.)
        Modified to scale with mean
        """
        nr, nc = self.x_mat.shape
        ncomp = self.scores.shape[1]
        A0 = 0 if type(self.x_mean) == int else 1
        ny = pd.np.sqrt(nr / (nr - ncomp - A0))
        E2 = self.x_mat*self.x_mat
        s = pd.np.sqrt(pd.np.nansum(E2, axis=1) / (nc - ncomp)) * ny
        S0 = pd.np.sqrt(pd.np.nansum(E2)/((nr - ncomp - A0) * (nc - ncomp)))
        return s/S0

    def plot(
        self,
        comps=['PC1', 'PC2'],
        classlevels=None,
        markers=None,
        classcolors=None,
        msize=100,
        figsize=(12, 8),
        plotpred=True,
        predsize=None,
        predlevels=None,
        predcolors=None,
        predmarkers=None,
        labels=None,
        predlabels=None,
        color='#555555'
    ):
        """Plot method for plotting scores, with optional classes and predictions"""
        return _plot(
            self, comps, classlevels, markers, classcolors, msize,
            figsize, plotpred, predsize, predlevels, predcolors, predmarkers,
            labels, predlabels, 10, color
        )

    def loadingsplot(
        self,
        comps=['PC1', 'PC2'],
        markers=None,
        color='0.7',
        msize=100,
        figsize=(12, 8),
        showweights=True,
        weightmarkers=None,
        weightcolors=None,
        weightsize=200,
        labels=True,
        textsize=10
    ):
        """Plot method for plotting loadings and optionally weights"""
        return _loadingsplot(
            modelinstance=self,
            comps=comps,
            markers=markers,
            color=color,
            msize=msize,
            figsize=figsize,
            showweights=showweights,
            weightmarkers=weightmarkers,
            weightcolors=weightcolors,
            weightsize=weightsize,
            labels=labels,
            textsize=textsize
        )

    def overviewplot(self):
        return pd.DataFrame(
            {
                'R2Y(cum)': self.R2Ycum,
                'Q2(cum)': getattr(self, 'Q2cum', None)
            },
            columns=['R2Y(cum)', 'Q2(cum)']
        ).plot(kind='bar', color=['g', 'b']).figure

    def dModXPlot(self):
        dmx = self.dModX()
        nr, nc = self.x_mat.shape
        ncomp = self.scores.shape[1]
        A0 = 0 if type(self.x_mean) == int else 1
        fc = pd.np.sqrt(f.isf(0.05, nr - ncomp - A0, nc - ncomp))
        ax = pd.Series(dmx, index=self.x_df.index.get_level_values(0)).plot(kind='bar', color='green')
        ax.hlines(fc, -1, 20)
        return ax.figure

    def dModYPlot(self):
        dmy = self.dModY()
        nr, nc = self.y_mat.shape
        ncomp = self.scores.shape[1]
        A0 = 0 if type(self.y_mean) == int else 1
        fc = pd.np.sqrt(f.isf(0.05, nr - ncomp - A0, nc - ncomp))
        ax = pd.Series(dmy, index=self.y_df.index.get_level_values(0)).plot(kind='bar', color='green')
        ax.hlines(fc, -1, 20)
        return ax.figure


class Nipals(object):
    """A Nipals class that can be used for PCA.

    Initialize with a Pandas DataFrame or an object that can be turned into a DataFrame
    (e.g. an array or a dict of lists)"""
    def __init__(self, x_df):
        super(Nipals, self).__init__()
        if type(x_df) != pd.core.frame.DataFrame:
            x_df = pd.DataFrame(x_df)
        self.x_df = x_df
        # Make sure data is numeric
        self.x_df = x_df.astype('float')
        # Check for and remove infs
        if pd.np.isinf(self.x_df).any().any():
            logging.warning("Data contained infinite values, converting to missing values")
            self.x_df.replace([pd.np.inf, -pd.np.inf], pd.np.nan, inplace=True)

    def _onecomp(self, mat, comp, hasna, startcol, tol, maxiter):
        nrt, nct = mat.shape
        miss = pd.np.isnan(mat)
        # Set t to column of X with highest var (uses original X mat)
        if startcol is None:
            xvar = pd.np.nanvar(self.x_mat, axis=0, ddof=1)
            startcol_use = pd.np.where(xvar == xvar.max())[0][0]
        else:
            startcol_use = startcol
        logging.info("PC {}, starting with column {}".format(comp, startcol_use))

        if hasna:
            mat_0 = pd.np.nan_to_num(mat)
            th = mat_0[:, startcol_use]
        else:
            th = mat[:, startcol_use]
        it = 0
        while True:
            # loadings
            if hasna:
                T2 = pd.np.repeat(th*th, nct)
                T2.shape = (nrt, nct)
                T2[miss] = 0
                ph = mat_0.T.dot(th) / T2.sum(axis=0)
            else:
                ph = mat.T.dot(th) / sum(th*th)
            # Normalize
            ph = ph / math.sqrt(pd.np.nansum(ph*ph))

            # Scores
            th_old = th
            if hasna:
                P2 = pd.np.repeat(ph*ph, nrt)
                P2.shape = (nct, nrt)
                P2[miss.T] = 0
                th = mat_0.dot(ph) / P2.sum(axis=0)
            else:
                th = mat.dot(ph) / sum(ph*ph)

            # Check convergence
            if pd.np.nansum((th-th_old)**2) < tol:
                break
            it += 1
            if it >= maxiter:
                raise RuntimeError(
                    "Convergence was not reached in {} iterations for component {}".format(
                        maxiter, comp
                    )
                )
        return th, ph

    def fit(
        self,
        ncomp=None,
        tol=0.000001,
        center=True,
        scale=True,
        maxiter=500,
        startcol=None,
        eigsweep=False,
        cv=False
    ):
        """The Fit method, will fit a PCA to the X data.

        Keyword arguments:
        ncomp - number of components, defaults to all
        tol - tolerance for convergence checking, defaults to 1E-6
        center - whether to center the data, defaults to True
        scale - whether to scale the data, defaults to True
        maxiter - maximum number of iterations before convergence is considered failed, defaults to 500
        startcol - column in X data to start iteration from, if set to None, the column with maximal variance is selected, defaults to None
        eigsweep - whether to sweep out eigenvalues from the final scores, defaults to False"""
        self.eigsweep = eigsweep
        if ncomp is None:
            ncomp = min(self.x_df.shape)
        elif ncomp > min(self.x_df.shape):
            ncomp = min(self.x_df.shape)
            logging.warning(
                'ncomp is larger than the max dimension of the x matrix.\n'
                'fit will only return {} components'.format(ncomp)
            )
        # Convert to np array
        self.x_mat = self.x_df.values
        if center:
            self.x_mean = pd.np.nanmean(self.x_mat, axis=0)
            self.x_mat = self.x_mat - self.x_mean
        else:
            self.x_mean = 0
        if scale:
            self.x_std = pd.np.nanstd(self.x_mat, axis=0, ddof=1)
            self.x_mat = self.x_mat / self.x_std
        else:
            self.x_std = 1

        TotalSS = pd.np.nansum(self.x_mat*self.x_mat)
        nr, nc = self.x_mat.shape
        # initialize outputs
        eig = pd.np.empty((ncomp,))
        R2cum = pd.np.empty((ncomp,))
        PRESS_SS = pd.np.empty((ncomp,))
        loadings = pd.np.empty((nc, ncomp))
        scores = pd.np.empty((nr, ncomp))

        # NA handling
        x_miss = pd.np.isnan(self.x_mat)
        hasna = x_miss.any()
        if hasna:
            logging.info("Data has NA values")

        # self.x_mat_0 = pd.np.nan_to_num(self.x_mat)
        # t = [None] * ncomp
        # p = [None] * ncomp
        self.eig = []
        if cv:
            if cv is True:
                cv = 7
            cvxn = int(pd.np.ceil(nr / cv))
            cvxgroups = pd.np.array(range(cvxn * cv)).reshape(cvxn, cv).T
            cvyn = int(pd.np.ceil(nc / cv))
            cvygroups = pd.np.array(range(cvyn * cv)).reshape(cvyn, cv).T
        else:
            cv = 0
        for comp in range(ncomp):
            # Matrixes to keep ps and ts from cv folds
            cvP = pd.np.empty((nr, nc))
            cvT = pd.np.empty((nr, nc))
            PRESS = 0
            # Calculate on full matrix
            th, ph = self._onecomp(self.x_mat, comp, hasna, startcol, tol, maxiter)
            for cvround in range(cv):
                train_mat = pd.np.delete(self.x_mat, cvxgroups[cvround], 0)
                _, ph_cv = self._onecomp(train_mat, comp, hasna, startcol, tol, maxiter)
                train_mat = pd.np.delete(self.x_mat, cvygroups[cvround], 1)
                th_cv, _ = self._onecomp(train_mat, comp, hasna, startcol, tol, maxiter)
                # Make sure the PCs are rotated in the same main direction for all cvs
                if pd.np.corrcoef(ph, ph_cv)[1, 0] < 0:
                    cvP[[grp for grp in cvxgroups[cvround] if grp < nr]] = -ph_cv
                else:
                    cvP[[grp for grp in cvxgroups[cvround] if grp < nr]] = ph_cv
                if pd.np.corrcoef(th, th_cv)[1, 0] < 0:
                    cvT.T[[grp for grp in cvygroups[cvround] if grp < nc]] = -th_cv
                else:
                    cvT.T[[grp for grp in cvygroups[cvround] if grp < nc]] = th_cv
            # Calculate PRESS
            if cv:
                pred_mat = cvP * cvT
                cv_res = self.x_mat - pred_mat
                cv_res[x_miss] = 0
                PRESS = pd.np.sum(cv_res**2)

            PRESS_SS[comp] = PRESS / pd.np.nansum(self.x_mat*self.x_mat)
            # Update X
            self.x_mat = self.x_mat - pd.np.outer(th, ph)
            loadings[:, comp] = ph
            scores[:, comp] = th
            eig[comp] = pd.np.nansum(th*th)

            # Cumulative proportion of variance explained
            R2cum[comp] = 1 - (pd.np.nansum(self.x_mat*self.x_mat) / TotalSS)

        # "Uncumulate" R2
        self.R2 = pd.np.insert(pd.np.diff(R2cum), 0, R2cum[0])
        self.R2cum = pd.Series(R2cum, index=["PC{}".format(i+1) for i in range(ncomp)])

        # Finalize eigenvalues and subtract from scores
        self.eig = pd.Series(pd.np.sqrt(eig))
        if self.eigsweep:
            scores = scores / self.eig.values
        if cv:
            self.PRESS_SS = pd.Series(PRESS_SS, index=["PC{}".format(i+1) for i in range(ncomp)])
            self.Q2 = 1 - self.PRESS_SS
            self.Q2cum = 1 - pd.np.cumprod(self.PRESS_SS)

        # Convert results to DataFrames
        self.scores = pd.DataFrame(scores, index=self.x_df.index, columns=["PC{}".format(i+1) for i in range(ncomp)])
        self.loadings = pd.DataFrame(loadings, index=self.x_df.columns, columns=["PC{}".format(i+1) for i in range(ncomp)])
        return True

    def dModX(self):
        """
        Calculates DModX for model, ported from pcaMethods
        (Stacklies W, Redestig H, Scholz M, Walther D and Selbig J (2007).
        “pcaMethods – a Bioconductor package providing PCA methods for incomplete
        data.” Bioinformatics, 23, pp. 1164–1167.)
        Modified to scale with mean
        """
        nr, nc = self.x_mat.shape
        ncomp = self.scores.shape[1]
        A0 = 0 if type(self.x_mean) == int else 1
        ny = pd.np.sqrt(nr / (nr - ncomp - A0))
        E2 = self.x_mat*self.x_mat
        s = pd.np.sqrt(pd.np.nansum(E2, axis=1) / (nc - ncomp)) * ny
        S0 = pd.np.sqrt(pd.np.nansum(E2)/((nr - ncomp - A0) * (nc - ncomp)))
        return s/S0

    def loadingsplot(
        self,
        comps=['PC1', 'PC2'],
        msize=100,
        figsize=(12, 8),
    ):
        """Plot method for plotting loadings"""
        return _loadingsplot(
            modelinstance=self,
            comps=comps,
            markers=None,
            color='0.7',
            msize=msize,
            figsize=figsize,
            showweights=False,
            weightmarkers=None,
            weightcolors=None,
            weightsize=None,
            labels=True,
            textsize=10
        )

    def plot(
        self,
        comps=['PC1', 'PC2'],
        classlevels=None,
        markers=None,
        classcolors=None,
        msize=100,
        figsize=(12, 8),
        plotpred=True,
        predsize=None,
        predlevels=None,
        predcolors=None,
        predmarkers=None,
        labels=None,
        predlabels=None,
        textsize=10,
        color='#555555'
    ):
        """Plot method for plotting scores, with optional classes and predictions"""
        return _plot(
            self, comps, classlevels, markers, classcolors, msize,
            figsize, plotpred, predsize, predlevels, predcolors, predmarkers,
            labels, predlabels, textsize, color
        )

    def predict(self, new_x):
        self.new_x = (new_x - self.x_mean) / self.x_std
        self.new_x = self.new_x.fillna(0)
        self.pred = self.new_x.dot(self.loadings)
        if self.eigsweep:
            self.pred = self.pred / self.eig.values
        return True

    def overviewplot(self):
        return pd.DataFrame(
            {
                'R2(cum)': self.R2cum,
                'Q2(cum)': getattr(self, 'Q2cum', None)
            },
            columns=['R2(cum)', 'Q2(cum)']
        ).plot(kind='bar', color=['g', 'b']).figure

    def dModXPlot(self):
        dmx = self.dModX()
        nr, nc = self.x_mat.shape
        ncomp = self.scores.shape[1]
        A0 = 0 if type(self.x_mean) == int else 1
        fc = pd.np.sqrt(f.isf(0.05, nr - ncomp - A0, nc - ncomp))
        ax = pd.Series(dmx, index=self.x_df.index.get_level_values(0)).plot(kind='bar', color='green')
        ax.hlines(fc, -1, 20)
        return ax.figure



import logging

import matplotlib
import pandas as pd
import pytest

from nipals import nipals

testdata = [
    [pd.np.nan, 67, 90, 98, 120],
    [pd.np.nan, 71, 93, 102, 129],
    [65, 76, 95, 105, 134],
    [50, 80, 102, 130, 138],
    [60, 82, 97, 135, 151],
    [65, 89, 106, 137, 153],
    [75, 95, 117, 133, 155]
]

testdata_full = [
    [50, 67, 90, 98, 120],
    [55, 71, 93, 102, 129],
    [65, 76, 95, 105, 134],
    [50, 80, 102, 130, 138],
    [60, 82, 97, 135, 151],
    [65, 89, 106, 137, 153],
    [75, 95, 117, 133, 155]
]

testdata_class = pd.DataFrame(testdata)
testdata_class['C1'] = [0, 0, 1, 1, 1, 0, 0]
testdata_class['C2'] = [1, 1, 0, 0, 0, 1, 1]
testdata_class = testdata_class.set_index(['C1', 'C2'], append=True)

yarn = pd.read_csv("tests/yarn.csv", index_col=0)
yarn_scores = pd.read_csv("tests/yarn_scores.csv", index_col=0).values
yarn_loadings = pd.read_csv("tests/yarn_loadings.csv", index_col=0).values
yarn_weights = pd.read_csv("tests/yarn_weights.csv", index_col=0).values
yarn_missing_scores = pd.read_csv("tests/yarn_missing_scores.csv", index_col=0, usecols=[0, 2, 3, 4]).values
yarn_missing_loadings = pd.read_csv("tests/yarn_missing_loadings.csv", index_col=0).values
yarn_missing_weights = pd.read_csv("tests/yarn_missing_weights.csv", index_col=0).values

oliveoil = pd.read_csv("tests/oliveoil.csv", index_col=0)
oliveoil_scores = pd.read_csv("tests/oliveoil_scores.csv", index_col=0)
oliveoil_missing_y = pd.read_csv("tests/oliveoil_miss.csv", index_col=0)
oliveoil_missing_y_scores = pd.read_csv("tests/oliveoil_miss_scores.csv", index_col=0)

# Set matplotlib backend to allow for headless testing
matplotlib.use('Agg')


def test_init_from_df():
    # It should be possible to init a nipals class from a Pandas DataFrame
    assert nipals.Nipals(pd.DataFrame(testdata)).__class__ == nipals.Nipals


def test_init_from_data():
    # It should be possible to init a nipals class from something that
    # can be made into a Pandas DataFrame
    assert nipals.Nipals(testdata).__class__ == nipals.Nipals


def test_run_pca():
    nip = nipals.Nipals(testdata)
    # Set startcol to get same results as R nipals:nipals (rounding errors give slightly different rotation otherwise)
    assert nip.fit(startcol=1)
    pd.np.testing.assert_almost_equal(list(nip.eig), [
        4.876206673116805,
        2.044242214135278,
        1.072810327696312,
        0.23701331933712338,
        0.14325051166784325
    ])
    # Also run without startcol set to make sure that it works as well. But compare to self data
    assert nip.fit()
    pd.np.testing.assert_almost_equal(list(nip.eig), [
        4.876216689582536,
        2.044275687396918,
        1.072805497059184,
        0.23696073749622645,
        0.14327789003413574
    ])


def test_call_with_too_large_ncomp(caplog):
    nip = nipals.Nipals(testdata)
    assert nip.fit(ncomp=10)
    assert caplog.record_tuples == [
        (
            'root',
            logging.WARNING,
            'ncomp is larger than the max dimension of the x matrix.\n'
            'fit will only return 5 components'
        ),
    ]


def test_run_pca_without_na():
    nip = nipals.Nipals(testdata_full)
    assert nip.fit()
    pd.np.testing.assert_almost_equal(list(nip.eig), [
        5.020518433605382,
        1.879323465996815,
        1.1081766447275905,
        0.17225187199265019,
        0.06936702860594454
    ])


def test_fail_from_maxiter():
    nip = nipals.Nipals(testdata_full)
    with pytest.raises(RuntimeError):
        nip.fit(maxiter=1)


def test_run_pca_with_set_ncomp():
    nip = nipals.Nipals(testdata_full)
    assert nip.fit(ncomp=2)
    pd.np.testing.assert_almost_equal(list(nip.eig), [5.020518433605, 1.879323465996])


def test_run_pca_with_precentered_data():
    centered = pd.DataFrame(testdata_full)
    centered = centered - centered.mean()
    nip = nipals.Nipals(centered)
    assert nip.fit(center=False, ncomp=2)
    pd.np.testing.assert_almost_equal(list(nip.eig), [5.020518433605, 1.879323465996])


def test_run_pca_with_prescaled_data():
    scaled = pd.DataFrame(testdata_full)
    scaled = (scaled - scaled.mean()) / scaled.std(ddof=1)
    nip = nipals.Nipals(scaled)
    assert nip.fit(center=False, scale=False, ncomp=2)
    pd.np.testing.assert_almost_equal(list(nip.eig), [5.020518433605, 1.879323465996])


def test_run_pca_check_scores_with_sweep():
    nip = nipals.Nipals(testdata)
    assert nip.fit(ncomp=2, eigsweep=True, startcol=1)
    pd.np.testing.assert_almost_equal(nip.scores.values, [
        [-0.5585132, 0.1224190],
        [-0.3801627, 0.1703718],
        [-0.2026926, 0.3163937],
        [-0.0337608, -0.6915786],
        [0.1285239, -0.4501362],
        [0.3562110, -0.2048250],
        [0.5982563, 0.3647261],
    ], 3)


def test_run_pca_check_scores():
    nip = nipals.Nipals(testdata)
    assert nip.fit(ncomp=2)
    pd.np.testing.assert_almost_equal(nip.scores.values, [
        [-2.72332498, 0.25021637],
        [-1.85369271, 0.34827344],
        [-0.98854112, 0.64658032],
        [-0.16429534, -1.41375825],
        [0.62686116, -0.92008715],
        [1.7370958, -0.41837244],
        [2.91721621, 0.746181]
    ])


def test_predict_from_pca():
    nip = nipals.Nipals(testdata)
    nip.fit(ncomp=2)
    assert nip.predict(pd.DataFrame([
        [63, 70, 98, 110, 124],
        [51, 82, 102, 110, 108]
    ]))
    pd.np.testing.assert_almost_equal(
        nip.pred.values,
        [
            [-1.4465766,  0.4500705],
            [-1.6229739, -0.340578]
        ]
    )


def test_predict_from_pca_with_sweep():
    nip = nipals.Nipals(testdata)
    nip.fit(ncomp=2, eigsweep=True)
    assert nip.predict(pd.DataFrame([
        [63, 70, 98, 110, 124],
        [51, 82, 102, 110, 108]
    ]))
    pd.np.testing.assert_almost_equal(
        nip.pred.values,
        [
            [-0.2966596, 0.2201614],
            [-0.3328347, -0.1666008]
        ]
    )


def test_plot():
    nip = nipals.Nipals(testdata)
    nip.fit(ncomp=2)
    plt = nip.plot()
    assert isinstance(plt, matplotlib.figure.Figure)
    return plt


def test_plot_classes():
    nip = nipals.Nipals(testdata_class)
    nip.fit(ncomp=2)
    plt = nip.plot(classlevels=['C1', 'C2'])
    assert isinstance(plt, matplotlib.figure.Figure)
    return plt


def test_plot_classoptions():
    nip = nipals.Nipals(testdata_class)
    nip.fit(ncomp=2)
    plt = nip.plot(classlevels=['C1', 'C2'], markers=['s', 'o'], classcolors=['red', 'black'])
    assert isinstance(plt, matplotlib.figure.Figure)
    return plt


def test_pred_plot():
    nip = nipals.Nipals(testdata_class)
    nip.fit(ncomp=2)
    nip.predict(pd.DataFrame([
        [63, 70, 98, 110, 124],
        [51, 82, 102, 110, 108]
    ]))
    plt = nip.plot(classlevels=['C1', 'C2'], plotpred=True)
    assert isinstance(plt, matplotlib.figure.Figure)
    return plt


def test_plot_labels():
    nip = nipals.Nipals(testdata_class)
    nip.fit(ncomp=2)
    nip.predict(pd.DataFrame([
        [63, 70, 98, 110, 124],
        [51, 82, 102, 110, 108]
    ], index=['a', 'b']))
    plt = nip.plot(classlevels=['C1', 'C2'], plotpred=True, labels='C1', predlabels=0)
    assert isinstance(plt, matplotlib.figure.Figure)
    return plt


def test_pred_plot_labels():
    nip = nipals.Nipals(testdata_class)
    nip.fit(ncomp=2)
    nip.predict(pd.DataFrame([
        [63, 70, 98, 110, 124, 0, 1, 0],
        [51, 82, 102, 110, 108, 1, 0, 0]
    ], index=['a', 'b']).set_index([5, 6, 7], append=True))
    plt = nip.plot(classlevels=['C1', 'C2'], plotpred=True, labels='C1', predlevels=[5, 6], predlabels=0)
    assert isinstance(plt, matplotlib.figure.Figure)
    return plt


def test_pred_plot_options():
    nip = nipals.Nipals(testdata_class)
    nip.fit(ncomp=2)
    nip.predict(pd.DataFrame([
        [63, 70, 98, 110, 124, 0, 1, 0],
        [51, 82, 102, 110, 108, 1, 0, 0]
    ]).set_index([5, 6, 7], append=True))
    with pytest.raises(KeyError):
        plt = nip.plot(classlevels=['C1', 'C2'], plotpred=True, predlevels=[5, 7])
    plt = nip.plot(classlevels=['C1', 'C2'], plotpred=True, predlevels=[5, 6])
    assert isinstance(plt, matplotlib.figure.Figure)
    return plt


def test_loadings_plot():
    nip = nipals.Nipals(testdata_class)
    nip.fit(ncomp=2)
    plt = nip.loadingsplot()
    assert isinstance(plt, matplotlib.figure.Figure)
    return plt


def test_pls():
    pls = nipals.PLS(yarn.iloc[:, :268], yarn.iloc[:, 268])
    assert pls.fit(ncomp=6, scale=False)
    pd.np.testing.assert_almost_equal(pls.scores.values, yarn_scores)
    pd.np.testing.assert_almost_equal(pls.loadings.values, yarn_loadings)
    pd.np.testing.assert_almost_equal(pls.weights.values, yarn_weights)


def test_pls_nocenter_and_scale():
    pls = nipals.PLS(yarn.iloc[:, :268] - yarn.iloc[:, :268].mean(), yarn.iloc[:, 268])
    assert pls.fit(ncomp=6, scale=False, center=False)
    pd.np.testing.assert_almost_equal(pls.scores.values, yarn_scores)
    pd.np.testing.assert_almost_equal(pls.loadings.values, yarn_loadings)
    pd.np.testing.assert_almost_equal(pls.weights.values, yarn_weights)


def test_pls_missing_x():
    tmpx = yarn.iloc[:, :268]
    csel = [171, 107, 222, 150, 76, 205, 63, 19, 121, 183]
    rsel = [23, 0, 3, 22, 15, 21, 19, 7, 19, 5]
    for r, c in zip(rsel, csel):
        tmpx.iloc[r, c] = pd.np.nan
    pls = nipals.PLS(tmpx, yarn.iloc[:, 268])
    assert pls.fit(ncomp=3)
    pd.np.testing.assert_almost_equal(pls.scores.values, yarn_missing_scores, 5)
    pd.np.testing.assert_almost_equal(pls.loadings.values, yarn_missing_loadings)
    pd.np.testing.assert_almost_equal(pls.weights.values, yarn_missing_weights)


def test_pls_nondf():
    pls = nipals.PLS(yarn.iloc[:, :268].values, yarn.iloc[:, 268].values)
    assert pls.fit(ncomp=6, scale=False)
    pd.np.testing.assert_almost_equal(pls.scores.values, yarn_scores)
    pd.np.testing.assert_almost_equal(pls.loadings.values, yarn_loadings)
    pd.np.testing.assert_almost_equal(pls.weights.values, yarn_weights)


def test_pls_optionvariations(caplog):
    pls = nipals.PLS(yarn.iloc[:, :268], yarn.iloc[:, 268])
    assert pls.fit()
    assert pls.fit(ncomp=500, startcol=0)
    assert caplog.record_tuples == [
        (
            'root',
            logging.WARNING,
            'ncomp is larger than the max dimension of the x matrix.\n'
            'fit will only return 28 components'
        ),
    ]
    with pytest.raises(RuntimeError):
        pls.fit(maxiter=1)


def test_pls_multiy():
    pls = nipals.PLS(oliveoil.iloc[:, :5], oliveoil.iloc[:, 5:])
    assert pls.fit(ncomp=2)
    pd.np.testing.assert_almost_equal(pls.scores.values, oliveoil_scores * [-1, 1], 4)


def test_pls_missing_y():
    pls = nipals.PLS(oliveoil_missing_y.iloc[:, :5], oliveoil_missing_y.iloc[:, 5:])
    assert pls.fit(ncomp=2)
    pd.np.testing.assert_almost_equal(pls.scores.values, oliveoil_missing_y_scores * [-1, 1], 3)


def test_pls_score_plot():
    pls = nipals.PLS(oliveoil_missing_y.iloc[:, :5], oliveoil_missing_y.iloc[:, 5:])
    assert pls.fit(ncomp=2)
    plt = pls.plot()
    assert isinstance(plt, matplotlib.figure.Figure)
    return plt


def test_pls_load_plot():
    pls = nipals.PLS(oliveoil_missing_y.iloc[:, :5], oliveoil_missing_y.iloc[:, 5:])
    assert pls.fit(ncomp=2)
    plt = pls.loadingsplot()
    assert isinstance(plt, matplotlib.figure.Figure)
    return plt


def test_pls_load_plot_no_y():
    pls = nipals.PLS(oliveoil_missing_y.iloc[:, :5], oliveoil_missing_y.iloc[:, 5:])
    assert pls.fit(ncomp=2)
    plt = pls.loadingsplot(showweights=False)
    assert isinstance(plt, matplotlib.figure.Figure)
    return plt


def test_pls_load_plot_without_labels():
    pls = nipals.PLS(oliveoil_missing_y.iloc[:, :5], oliveoil_missing_y.iloc[:, 5:])
    assert pls.fit(ncomp=2)
    plt = pls.loadingsplot(labels=False)
    assert isinstance(plt, matplotlib.figure.Figure)
    return plt


def test_pls_load_plot_with_markers():
    pls = nipals.PLS(oliveoil_missing_y.iloc[:, :5], oliveoil_missing_y.iloc[:, 5:])
    assert pls.fit(ncomp=2)
    plt = pls.loadingsplot(weightmarkers=['s', 'o', 'v', '^', '+', '3'])
    assert isinstance(plt, matplotlib.figure.Figure)
    return plt


def test_inf_values_in_dfs(caplog):
    olive_inf = oliveoil.replace([18.7, 75.1], pd.np.inf)
    nipals.Nipals(olive_inf)
    nipals.PLS(olive_inf.iloc[:, :5], olive_inf.iloc[:, 5:])
    assert caplog.record_tuples == [
        (
            'root',
            logging.WARNING,
            'Data contained infinite values, converting to missing values'
        ),
        (
            'root',
            logging.WARNING,
            'X data contained infinite values, converting to missing values'
        ),
        (
            'root',
            logging.WARNING,
            'Y data contained infinite values, converting to missing values'
        ),
    ]


def test_cv_pca():
    nip = nipals.Nipals(testdata_class)
    assert nip.fit(ncomp=2, cv=True)
    assert nip.fit(ncomp=2, cv=4)
    pd.np.testing.assert_almost_equal(
        nip.R2,
        [0.8112004, 0.144992]
    )
    pd.np.testing.assert_almost_equal(
        nip.Q2,
        [0.5779837, 0.1907985]
    )


def test_cv_pca_with_fliped_axis():
    """Checks that the Q2 is approx the same when some cvs have flipped axis in some PC"""
    nip = nipals.Nipals(testdata_class)
    assert nip.fit(ncomp=2, cv=True, startcol=2)
    Q2_sc2 = nip.Q2.values
    assert nip.fit(ncomp=2, cv=True)
    pd.np.testing.assert_almost_equal(
        nip.Q2,
        Q2_sc2,
        3
    )


def test_cv_pls():
    pls = nipals.PLS(oliveoil.iloc[:, :5], oliveoil.iloc[:, 5:])
    assert pls.fit(ncomp=2, cv=True)
    assert pls.fit(ncomp=2, cv=4)
    pd.np.testing.assert_almost_equal(
        pls.Q2,
        [0.2684853, 0.0644187]
    )
    pd.np.testing.assert_almost_equal(
        pls.R2X,
        [0.5826444, 0.2367458]
    )
    pd.np.testing.assert_almost_equal(
        pls.R2Y,
        [0.4326835, 0.0856207]
    )


def test_dmodx_pca():
    nip = nipals.Nipals(testdata_class)
    assert nip.fit(ncomp=2)
    pd.np.testing.assert_almost_equal(
        nip.dModX(),
        [
            0.5364634, 0.3480396, 0.5702191,
            1.5525099, 1.5752593, 0.3275259,
            1.1255771
        ]
    )


def test_dmodx_plot_pca():
    nip = nipals.Nipals(testdata_class)
    assert nip.fit(ncomp=2)
    plt = nip.dModXPlot()
    assert isinstance(plt, matplotlib.figure.Figure)
    return plt


def test_overviewplotplot_pca():
    nip = nipals.Nipals(testdata_class)
    assert nip.fit(ncomp=2)
    plt = nip.overviewplot()
    assert isinstance(plt, matplotlib.figure.Figure)
    return plt


def test_dmod_pls():
    pls = nipals.PLS(oliveoil.iloc[:, :5], oliveoil.iloc[:, 5:])
    assert pls.fit(ncomp=2)
    pd.np.testing.assert_almost_equal(
        pls.dModX(),
        [
            0.8561775, 0.6808639, 1.2475466, 1.9522224,
            0.1694258, 0.6151471, 0.8651022, 1.1295028,
            1.2212977, 1.2945167, 0.8681262, 1.141947,
            0.7637504, 0.3900809, 0.7382779, 0.7063927
        ]
    )
    pd.np.testing.assert_almost_equal(
        pls.dModY(),
        [
            0.5728738, 2.0598601, 1.2420525, 0.5257193, 1.4290988, 0.7752674,
            1.0029673, 0.8943648, 1.0509669, 0.8511583, 0.7153933, 1.0285289,
            0.8238886, 0.4616424, 0.5664762, 0.7409856
        ]
    )


def test_dmodx_plot_pls():
    pls = nipals.PLS(oliveoil.iloc[:, :5], oliveoil.iloc[:, 5:])
    assert pls.fit(ncomp=2)
    plt = pls.dModXPlot()
    assert isinstance(plt, matplotlib.figure.Figure)
    return plt


def test_dmody_plot_pls():
    pls = nipals.PLS(oliveoil.iloc[:, :5], oliveoil.iloc[:, 5:])
    assert pls.fit(ncomp=2)
    plt = pls.dModYPlot()
    assert isinstance(plt, matplotlib.figure.Figure)
    return plt


def test_overviewplotplot_pls():
    pls = nipals.PLS(oliveoil.iloc[:, :5], oliveoil.iloc[:, 5:])
    assert pls.fit(ncomp=2)
    plt = pls.overviewplot()
    assert isinstance(plt, matplotlib.figure.Figure)
    return plt



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function

import io
import re
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import find_packages
from setuptools import setup


def read(*names, **kwargs):
    return io.open(
        join(dirname(__file__), *names),
        encoding=kwargs.get('encoding', 'utf8')
    ).read()


setup(
    name='nipals',
    version='0.4.1',
    license='MIT license',
    description='A module for calculation of PCA with the NIPALS algorithm',
    long_description='%s\n%s' % (
        re.compile('^.. start-badges.*^.. end-badges', re.M | re.S).sub('', read('README.rst')),
        re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', read('CHANGELOG.rst'))
    ),
    author='Fredrik Wallner',
    author_email='fredrik@wallner.nu',
    url='https://github.com/fredrikw/python-nipals',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        # 'Development Status :: 5 - Production/Stable',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        # 'Programming Language :: Python :: 3.3',
        # 'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        # 'Programming Language :: Python :: Implementation :: PyPy',
        # uncomment if you test on these interpreters:
        # 'Programming Language :: Python :: Implementation :: IronPython',
        # 'Programming Language :: Python :: Implementation :: Jython',
        # 'Programming Language :: Python :: Implementation :: Stackless',
        'Topic :: Utilities',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Chemistry',
    ],
    keywords=[
        # eg: 'keyword1', 'keyword2', 'keyword3',
    ],
    install_requires=[
        'pandas>=0.22',
        'scipy>=1.0.0',
        'matplotlib>2.0.0'
        # eg: 'aspectlib==1.1.1', 'six>=1.7',
    ],
    extras_require={
        # eg:
        #   'rst': ['docutils>=0.11'],
        #   ':python_version=="2.6"': ['argparse'],
    },
)
